/*=============================================================================
	Generic systems code
	By Higor, feel free to use this code on your project.
=============================================================================*/

#include "XC_Engine.h"

// Level watcher, performs various tasks on the loaded level
#include "UnXC_LevelWatcher.h"
FXC_LevelWatcher::FXC_LevelWatcher() {}

UBOOL FXC_LevelWatcher::Init()
{
	Level = NULL;
	OldACUnique = 0;
	CurActorNameIdx = 0;
	return true;
}

struct ActorChannelPair
{
	INT Hash;
	AActor* Actor;
	UActorChannel* Channel;
};

INT FXC_LevelWatcher::Tick( FLOAT DeltaSeconds)
{
	guard(FXC_LevelWatcher::Tick);

	TickCount++;
	//Level change detector
	guard(LevelCheck);
	ULevel* ALevel = (&Engine->GLevel)[Engine->b451Setup];
	if ( ALevel != Level )
	{
		ULevel* AEntry = (&Engine->GEntry)[Engine->b451Setup];
		Level = (ALevel != AEntry) ? ALevel : NULL;
		//Level changed
		if ( Level )
		{
			CurActorNameIdx = Level->iFirstDynamicActor;
		}
	}
	if ( !Level )
		return 0;
	unguard;
	
	//Client/Demo network play channel name normalization
	if ( bHandlesAcquired && (OldACUnique != UActorChannel::StaticClass()->ClassUnique) )
	{
		//Generically select up to 2 connections
		INT iConn = 0;
		UXC_NetConnectionHack* Conn[2];
		if ( Level->NetDriver && Level->NetDriver->ServerConnection )
			Conn[iConn++] = (UXC_NetConnectionHack*)Level->NetDriver->ServerConnection;
		if ( Level->DemoRecDriver && (Level->NetDriver != Level->DemoRecDriver) )
		{
			if ( Level->DemoRecDriver->ServerConnection )
				Conn[iConn++] = (UXC_NetConnectionHack*)Level->DemoRecDriver->ServerConnection;
			else if ( Level->DemoRecDriver->ClientConnections.Num() > 0 )
				Conn[iConn++] = (UXC_NetConnectionHack*)Level->DemoRecDriver->ClientConnections(0);
		}

		//Start renaming
		guard(RenameChannels);
		for ( INT i=0 ; i<iConn ; i++ )
		{
			TArray<ActorChannelPair>* AC = (TArray<ActorChannelPair>*) Conn[i]->GetActorChannels();
			for ( INT k=0 ; k<AC->Num() ; k++ )
			{
				UActorChannel* Ch = (*AC)(k).Channel;
				if ( Ch && (Ch->GetOuter() != Conn[i]) )
				{
					(Ch->*UnhashObjectFunc)( Ch->GetOuter() ? Ch->GetOuter()->GetIndex() : 0 );
					TCHAR sName[17];
					appSprintf( sName, TEXT("ActorChannel%i"), Ch->ChIndex);
					FName NewName = FName(sName);
					*(FName*) (DWORD(Ch) + 32) = NewName; //Rename
					*(UObject**) (DWORD(Ch) + 24) = Conn[i]; //Change outer
					(Ch->*HashObjectFunc)();
				}
			}
		}
		unguard;

		//Connected to a server, keep actor channel names low
		if ( UActorChannel::StaticClass()->ClassUnique > 500 + 1024 * (1-bHandlesAcquired) )
			UActorChannel::StaticClass()->ClassUnique = 0;
		OldACUnique = UActorChannel::StaticClass()->ClassUnique;
	}
	
	//Find bNetTemporary actors that could use a name reset
	guard(ActorNames);
	AActor* Actor = NULL;
	if ( CurActorNameIdx >= Level->Actors.Num() )
		CurActorNameIdx = Level->iFirstDynamicActor;
	else if ( TickCount % 8 > 0 )
	{
		INT Id = Max(Level->Actors.Num() - (INT)(TickCount%8), Level->iFirstDynamicActor);
		if ( Id < Level->Actors.Num() )
		Actor = Level->Actors( Id); //Query 7 out of 8 times
	}
	else
		Actor = Level->Actors(CurActorNameIdx++);
	if ( Actor && Level->NetDriver && Level->NetDriver->ServerConnection )
	{
		const TCHAR* cName = Actor->GetClass()->GetName();
		INT NameNumberIdx = appStrlen( cName );
		while ( (NameNumberIdx > 0) && appIsDigit(cName[NameNumberIdx-1]) )
			NameNumberIdx--;
		TCHAR BaseName[64];
		appStrncpy( BaseName, cName, NameNumberIdx+1);
		//These actors are good candidates for aggresive cycle sweeping
//		if ( Actor->LifeSpan != 0 ) 
		{
			//Negate if we find ONE name with at least 2 digits in name (less than 100 instances)
			if ( (Actor->GetClass()->ClassUnique > 99) && (appStrlen( Actor->GetName()) > NameNumberIdx+2) )
			{
				UBOOL bReduce = true;
				for ( INT i=Level->iFirstDynamicActor ; i<Level->Actors.Num() ; i++ )
				{
					AActor* AA = Level->Actors(i);
					if ( AA && (AA->GetClass() == Actor->GetClass()) && (appStrlen(AA->GetName()) <= NameNumberIdx+2) )
					{
						bReduce = false;
						break;
					}
				}
				if ( bReduce )
				{
					Actor->GetClass()->ClassUnique = 0;
				//	debugf( NAME_Log, TEXT("Reset names for %s"), cName);
				}
			}
		}
	}
	unguard;
	
	return 1;
	unguard;
}

UBOOL FXC_LevelWatcher::IsTyped( const TCHAR* Type)
{
	return appStricmp( Type, TEXT("LevelWatcher")) == 0;
}



// Halfassed time manager, used to correct game tick.
#include "UnXC_TimeManager.h"

FXC_TimeManager::FXC_TimeManager()
:	MSecAccumulator(0)
,	MSecInterval(100)
,	Factor(1)
,	DeltaAcc(0)
,	IgnoreTimers(10)
{
	appMemzero( &SystemDigest[0], 80); //Size of both arrays

	INT Year;
	INT Month;
	INT DayOfWeek;
	INT Day;
	INT Hour;
	INT Min;
	appSystemTime( Year, Month, DayOfWeek, Day, Hour, Min, LastSec, LastMSec );
}

UBOOL FXC_TimeManager::IsTyped( const TCHAR* Type)
{
	return appStricmp( Type, TEXT("TimeManager")) == 0;
}

INT FXC_TimeManager::Tick( FLOAT DeltaSeconds)
{
	INT Year;
	INT Month;
	INT DayOfWeek;
	INT Day;
	INT Hour;
	INT Min;
	INT Sec;
	INT MSec;
	INT i=0;

	appSystemTime( Year, Month, DayOfWeek, Day, Hour, Min, Sec, MSec );

	if ( Sec < LastSec )
		LastSec -= 60;

	i = MSec - LastMSec;
	if ( (Sec - LastSec) % 60 )
		i += ((Sec - LastSec) % 60) * 1000;
	MSecAccumulator += i;
	DeltaAcc += DeltaSeconds;
	if ( MSecAccumulator >= MSecInterval ) //DO THE MATH HERE
	{
//		if ( MSecAccumulator >= MSecInterval * 2 )
//			debugf( NAME_Log, TEXT("MSEC: SYSTEM RESOURCES HOGGED") );
	
		appMemmove( &SystemDigest[1], &SystemDigest[0], 36 );
		appMemmove( &TimeDigest[1], &TimeDigest[0], 36 );
		SystemDigest[0] = MSec;
		while ( SystemDigest[0] < SystemDigest[1] )
			SystemDigest[0] += 1000;
		TimeDigest[0] = TimeDigest[1] + DeltaAcc;
		for ( i=0 ; i<10 ; i++ )
			TimeDigest[i] -= TimeDigest[9];

		if ( SystemDigest[9] > 1000 )
			for ( i=0 ; i<10 ; i++ )
				SystemDigest[i] -= 1000;

//		INT AvgMSec = (SystemDigest[0] - SystemDigest[9]) / 9;
		INT DevMSec = 0;
		for ( i=1 ; i<10 ; i++ )
			DevMSec += Abs(((SystemDigest[9 - i] - SystemDigest[9]) / i) - MSecInterval);

		if ( DevMSec < MSecInterval )
		{
			FLOAT Delta = TimeDigest[0] * 1000;
//			debugf( NAME_Log, TEXT("MSEC TIMER %i %i %i %f"), MSecAccumulator, AvgMSec, DevMSec, Delta );
			FLOAT fMSecInv = _Reciprocal( (FLOAT)(SystemDigest[0] - SystemDigest[9]) );
			if ( IgnoreTimers <= 0 ) //Don't change factor after ignore timers expire
				Factor = _Reciprocal(Delta * fMSecInv);
			else
				IgnoreTimers--;
		}
//		else
//			debugf( NAME_Log, TEXT("MSEC: DEVIATION TOO HIGH  %i %i %i"), MSecAccumulator, AvgMSec, DevMSec );

		MSecAccumulator = MSecAccumulator % MSecInterval;
		DeltaAcc = 0;
	}
	
	LastMSec = MSec;
	LastSec = Sec;
	return 1;
}



UBOOL FXC_TimeManager::Exec( const TCHAR* Cmd, FOutputDevice& Ar )
{
	guard(FXC_TimeManager::Exec);
	if ( ParseCommand(&Cmd,TEXT("TimeFactor")) )
		Ar.Log( * FString::Printf( TEXT("XC_TimeManager time scaler is set at Factor=%f"), Factor) );
	else
		return 0;
	return 1;
	unguard;
}